"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const t = require("babel-types");
const utils_1 = require("./utils");
const jsx_1 = require("./jsx");
const constant_1 = require("./constant");
const create_html_element_1 = require("./create-html-element");
const babel_generator_1 = require("babel-generator");
const render_1 = require("./render");
const lodash_1 = require("lodash");
function buildConstructor() {
    const ctor = t.classMethod('constructor', t.identifier('constructor'), [], t.blockStatement([
        t.expressionStatement(t.callExpression(t.identifier('super'), [
            t.spreadElement(t.identifier('arguments'))
        ]))
    ]));
    return ctor;
}
function resetThisState() {
    return t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.thisExpression(), t.identifier('state')), t.callExpression(t.memberExpression(t.thisExpression(), t.identifier('_createData')), [])));
}
function isArrayMapCallExpression(callExpression) {
    if (callExpression &&
        t.isCallExpression(callExpression.node) &&
        t.isMemberExpression(callExpression.node.callee) &&
        t.isIdentifier(callExpression.node.callee.property, { name: 'map' })) {
        return true;
    }
    return false;
}
class Transformer {
    constructor(path, isRoot, componentSourceMap, location) {
        this.result = {
            template: '',
            components: []
        };
        this.methods = new Map();
        this.initState = new Set();
        this.jsxReferencedIdentifiers = new Set();
        this.customComponents = new Map();
        this.loopComponents = new Map();
        this.renderMethod = null;
        this.duplicateComponents = new Map();
        this.customComponentNames = new Set();
        this.classPath = path;
        this.isRoot = isRoot;
        this.componentSourceMap = componentSourceMap;
        this.location = location;
        this.moduleNames = Object.keys(path.scope.getAllBindings('module'));
        this.compile();
    }
    traverse() {
        const self = this;
        self.classPath.traverse({
            ClassMethod(path) {
                const node = path.node;
                if (t.isIdentifier(node.key)) {
                    const name = node.key.name;
                    self.methods.set(name, path);
                    if (name === 'render') {
                        self.renderMethod = path;
                    }
                    if (name === 'constructor') {
                        path.traverse({
                            AssignmentExpression(p) {
                                if (t.isMemberExpression(p.node.left) &&
                                    t.isThisExpression(p.node.left.object) &&
                                    t.isIdentifier(p.node.left.property) &&
                                    p.node.left.property.name === 'state' &&
                                    t.isObjectExpression(p.node.right)) {
                                    const properties = p.node.right.properties;
                                    properties.forEach(p => {
                                        if (t.isObjectProperty(p)) {
                                            if (t.isIdentifier(p.key)) {
                                                self.initState.add(p.key.name);
                                            }
                                        }
                                    });
                                }
                            }
                        });
                    }
                }
            },
            ClassProperty(path) {
                const { key: { name }, value } = path.node;
                if (t.isArrowFunctionExpression(value) || t.isFunctionExpression(value)) {
                    self.methods.set(name, path);
                }
                if (name === 'state' && t.isObjectExpression(value)) {
                    value.properties.forEach(p => {
                        if (t.isObjectProperty(p)) {
                            if (t.isIdentifier(p.key)) {
                                self.initState.add(p.key.name);
                            }
                        }
                    });
                }
            },
            JSXElement(path) {
                const id = path.node.openingElement.name;
                if (t.isJSXIdentifier(id) &&
                    !constant_1.DEFAULT_Component_SET.has(id.name) &&
                    self.moduleNames.indexOf(id.name) !== -1) {
                    const callExpression = path.findParent(p => p.isCallExpression());
                    if (isArrayMapCallExpression(callExpression)) {
                        let parentCallExpression = callExpression.findParent(p => isArrayMapCallExpression(p));
                        if (parentCallExpression) {
                            while (true) {
                                const temp = parentCallExpression.findParent(p => isArrayMapCallExpression(p));
                                if (!temp) {
                                    break;
                                }
                                parentCallExpression = temp;
                            }
                            const loopComponent = {
                                parent: callExpression,
                                element: path,
                                name: id.name
                            };
                            if (self.loopComponents.has(parentCallExpression)) {
                                const loopComponents = self.loopComponents.get(parentCallExpression);
                                self.loopComponents.set(parentCallExpression, loopComponents.concat(loopComponent));
                            }
                            else {
                                self.loopComponents.set(parentCallExpression, [loopComponent]);
                            }
                            // self.subLoopComponents.set(path, id.name)
                        }
                        else {
                            const loopComponent = {
                                parent: null,
                                element: path,
                                name: id.name
                            };
                            if (self.loopComponents.has(callExpression)) {
                                const loopComponents = self.loopComponents.get(callExpression);
                                self.loopComponents.set(callExpression, loopComponents.concat(loopComponent));
                            }
                            else {
                                self.loopComponents.set(callExpression, [loopComponent]);
                            }
                            // self.topLoopComponents.set(path, id.name)
                        }
                    }
                    else {
                        self.customComponents.set(path, id.name);
                    }
                }
            }
        });
    }
    renameImportJSXElement() {
        this.customComponents.forEach((name, path) => {
            if (this.duplicateComponents.has(name)) {
                const times = this.duplicateComponents.get(name);
                this.duplicateComponents.set(name, times + 1);
                this.customComponents.set(path, `${name}_${times + 1}`);
            }
            else {
                this.duplicateComponents.set(name, 0);
            }
        });
    }
    setProps() {
        const body = this.classPath.node.body.body;
        this.classPath.node.body.body = [
            build$PropsProperty(this.customComponents, body, this.jsxReferencedIdentifiers, this.renderMethod.get('body').scope),
            build$ComponentsProperty(this.customComponents, this.loopComponents),
            ...body
        ];
    }
    handleLoopComponents() {
        const uid = utils_1.incrementId();
        if (this.loopComponents.size > 0) {
            const properties = [];
            this.loopComponents.forEach((loopComponents, rootCallExpression) => {
                let iterator;
                let index;
                let blockStatement = [];
                const callee = rootCallExpression.node.callee;
                const calleeCode = babel_generator_1.default(callee).code;
                let nodes = [];
                loopComponents.forEach((loopComponent) => {
                    const { name, element: component, parent } = loopComponent;
                    let subscript = '';
                    const callExpression = component.findParent(p => isArrayMapCallExpression(p));
                    const mapFunction = callExpression.node.arguments[0];
                    if (t.isFunctionExpression(mapFunction) || t.isArrowFunctionExpression(mapFunction)) {
                        iterator = mapFunction.params[0];
                        index = mapFunction.params[1];
                    }
                    // @TODO: children 同级的情况
                    if (parent) {
                        const callee = callExpression.node.callee;
                        const calleeCode = babel_generator_1.default(callee).code;
                        subscript = calleeCode.split('.').slice(1, calleeCode.split('.').length).join('');
                    }
                    const node = this.generateTopLoopNodes(name, component, uid, subscript.slice(0, subscript.length - 3), parent, iterator.name, index);
                    nodes.push(node);
                });
                let stateName = '';
                if (!isContainThis(rootCallExpression.node.callee)) {
                    stateName = calleeCode.slice(0, calleeCode.length - 4);
                }
                else {
                    // todo 找倒数第二个 callee id
                    let ary = calleeCode.split('.');
                    stateName = ary[1] === 'state' || ary[1] === 'props' || ary[1] === '__state'
                        ? ary[2] : ary[1];
                }
                const returnStatement = t.returnStatement(t.objectExpression([
                    t.objectProperty(t.identifier('stateName'), t.stringLiteral(stateName)),
                    t.objectProperty(t.identifier('loopComponents'), t.callExpression(t.identifier(constant_1.INTERNAL_DYNAMIC), [
                        t.thisExpression(), t.identifier('nodes'), buildInternalSafeGet(stateName), t.stringLiteral(stateName)
                    ]))
                ]));
                const nodeDeclare = t.variableDeclaration('const', [
                    t.variableDeclarator(t.identifier('nodes'), t.arrayExpression(nodes))
                ]);
                blockStatement.push(nodeDeclare);
                blockStatement.push(returnStatement);
                properties.push(
                // t.objectMethod('method', t.identifier(createUUID()), [], t.blockStatement(blockStatement)),
                t.objectProperty(t.identifier(utils_1.createUUID()), t.arrowFunctionExpression([], t.blockStatement(blockStatement))));
            });
            this.classPath.node.body.body.unshift(t.classProperty(t.identifier('$dynamicComponents'), t.objectExpression(properties)));
        }
    }
    generateTopLoopNodes(name, component, uid, subscript, parent, iterator, index) {
        const properties = [];
        const self = this;
        component.traverse({
            JSXAttribute(path) {
                const attr = path.node;
                const name = attr.name;
                let value;
                if (t.isJSXElement(attr.value)) {
                    throw utils_1.codeFrameError(attr.loc, 'JSX 参数不支持传入 JSX 元素');
                }
                else if (t.isJSXExpressionContainer(attr.value)) {
                    const expresion = attr.value.expression;
                    const expresionPath = path.get('value').get('expression');
                    if (t.isIdentifier(expresion)) {
                        // TODO: 查看当前 value 的作用域
                        let replacement = t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), expresion);
                        if (expresion.name === iterator) {
                            replacement = t.identifier(constant_1.MAP_CALL_ITERATOR);
                        }
                        else if (expresion.name === index) {
                            replacement = t.identifier(index);
                        }
                        expresionPath.replaceWith(replacement);
                    }
                    else if (t.isMemberExpression(expresion)) {
                        if (!t.isThisExpression(expresion.object) && !isContainThis(expresion)) {
                            let replacement = buildInternalSafeGet(babel_generator_1.default(expresion).code, true);
                            const id = utils_1.findFirstIdentifierFromMemberExpression(expresion);
                            if (id.name === iterator) {
                                id.name = constant_1.MAP_CALL_ITERATOR;
                                replacement = expresion;
                            }
                            expresionPath.replaceWith(replacement);
                        }
                        else if (isNeedClassMethodWrapper(name.name, expresion)) {
                            const id = `$$anonymousPropsFunction_${uid()}`;
                            path.get('value').get('expression').replaceWith(t.memberExpression(t.thisExpression(), t.identifier(id)));
                            self.classPath.node.body.body.push(buildAnonymousClassMethod(expresion, id));
                        }
                    }
                    else {
                        expresionPath.traverse({
                            Identifier(path) {
                                const { parent, node } = path;
                                if (!t.isMemberExpression(parent) && node.name !== constant_1.INTERNAL_SAFE_GET) {
                                    let replacement = t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), node);
                                    if (node.name === iterator) {
                                        replacement = t.memberExpression(t.identifier(constant_1.MAP_CALL_ITERATOR), t.identifier(iterator));
                                    }
                                    else if (node.name === index) {
                                        replacement = t.memberExpression();
                                    }
                                    path.replaceWith(replacement);
                                }
                            },
                            MemberExpression(path) {
                                const { parent, node } = path;
                                if (!isContainThis(node)) {
                                    let replacement = node;
                                    const id = utils_1.findFirstIdentifierFromMemberExpression(node);
                                    if (id.name === iterator) {
                                        id.name = constant_1.MAP_CALL_ITERATOR;
                                        replacement = node;
                                        // console.log('fuck')
                                        // node.object = id
                                    }
                                    if (parent !== expresion && path.scope.hasBinding(utils_1.findFirstIdentifierFromMemberExpression(replacement).name)) {
                                        replacement = buildInternalSafeGet(babel_generator_1.default(node).code);
                                    }
                                    if (replacement !== node) {
                                        path.replaceWith(replacement);
                                    }
                                }
                            }
                        });
                    }
                    value = attr.value.expression;
                }
                else {
                    value = attr.value;
                }
                properties.push(t.objectProperty(t.identifier(name.name), value));
            }
        });
        const argsFunction = t.objectMethod('method', t.identifier('args'), [t.identifier(constant_1.MAP_CALL_ITERATOR), t.identifier('index')], t.blockStatement([
            t.returnStatement(t.objectExpression(properties))
        ]));
        const objectName = t.objectProperty(t.identifier('name'), t.stringLiteral(name));
        const objectPath = t.objectProperty(t.identifier('path'), t.identifier(name));
        const objectSubscript = t.objectProperty(t.identifier('subscript'), t.stringLiteral(subscript));
        const object = t.objectExpression([
            objectName,
            objectPath,
            objectSubscript,
            argsFunction
        ]);
        return parent
            ? t.objectExpression([t.objectProperty(t.identifier('children'), t.arrayExpression([object]))])
            : object;
    }
    setComponentResult() {
        this.componentSourceMap.forEach((names, source) => {
            let name = '';
            names.forEach(n => {
                this.loopComponents.forEach((lc) => {
                    lc.forEach(c => {
                        if (names.includes(c.name)) {
                            name = names[0];
                        }
                    });
                });
                if (this.duplicateComponents.has(n)) {
                    name = n;
                }
            });
            if (name) {
                this.result.components.push({
                    name,
                    path: source
                });
            }
        });
        this.result.components = lodash_1.uniqBy(this.result.components, 'name');
    }
    replaceImportedJSXElement() {
        this.customComponents.forEach((name, path) => {
            path.replaceWith(jsx_1.buildRefTemplate(findImportedName(name), name));
            this.customComponentNames.add(name);
        });
        this.loopComponents.forEach(lc => {
            lc.forEach(c => {
                const callExpression = c.element.find(p => isArrayMapCallExpression(p));
                const [func] = callExpression.node.arguments;
                if (t.isFunctionExpression(func) ||
                    t.isArrowFunctionExpression(func)) {
                    const [item] = func.params;
                    if (t.isIdentifier(item)) {
                        c.element.replaceWith(jsx_1.buildRefTemplate(findImportedName(c.name), item.name, true));
                    }
                }
            });
        });
        this.result.template = this.result.components.reduce((acc, { name, path }, index) => {
            return acc + `${index === 0 ? '' : `\n`}` + create_html_element_1.createHTMLElement({
                name: 'import',
                attributes: {
                    src: utils_1.pathResolver(path, this.location)
                },
                value: ''
            });
        }, '');
    }
    resetConstructor() {
        if (this.methods.has('constructor')) {
            const ctor = this.methods.get('constructor');
            if (ctor && ctor.isClassMethod()) {
                ctor.node.body.body.push(resetThisState());
            }
        }
        else {
            const ctor = buildConstructor();
            ctor.body.body.push(resetThisState());
            this.classPath.node.body.body.unshift(ctor);
        }
    }
    parseRender() {
        if (this.renderMethod) {
            const instanceName = this.classPath.node.id.name;
            this.result.template = this.result.template
                + render_1.parseRender(this.renderMethod, this.methods, this.initState, this.isRoot, instanceName, this.jsxReferencedIdentifiers, this.customComponentNames);
        }
    }
    compile() {
        this.traverse();
        this.handleLoopComponents();
        this.renameImportJSXElement();
        this.setProps();
        this.setComponentResult();
        this.replaceImportedJSXElement();
        this.resetConstructor();
        this.parseRender();
    }
}
exports.Transformer = Transformer;
function buildInternalSafeGet(getter, isData) {
    return t.callExpression(t.identifier(constant_1.INTERNAL_SAFE_GET), [
        !isData ? t.memberExpression(t.thisExpression(), t.identifier('state')) : t.identifier(constant_1.MAP_CALL_ITERATOR),
        t.stringLiteral(getter)
    ]);
}
function isNeedClassMethodWrapper(name, expresion) {
    if (name.startsWith('on')) {
        const ids = babel_generator_1.default(expresion).code.split('.');
        if (ids[0] === 'this' && ids[1] === 'props' && ids[2] && ids[3] !== 'bind') {
            return true;
        }
    }
    return false;
}
function isContainThis(expresion) {
    const ids = babel_generator_1.default(expresion).code.split('.');
    if (ids[0] === 'this') {
        return true;
    }
    return false;
}
function buildAnonymousClassMethod(expresion, id) {
    return t.classMethod('method', t.identifier(id), [], t.blockStatement([
        t.expressionStatement(t.callExpression(expresion, [
            t.spreadElement(t.identifier('arguments'))
        ]))
    ]));
}
function build$PropsProperty(importedJSXElement, body, jsxReferencedIdentifiers, renderScope) {
    const uid = utils_1.incrementId();
    const properties = [];
    importedJSXElement.forEach((name, path) => {
        let attrObj = [t.objectProperty(t.identifier('$name'), t.stringLiteral(name))];
        path.traverse({
            Identifier(path) {
                const parentPath = path.parentPath;
                if (parentPath.isConditionalExpression() ||
                    parentPath.isLogicalExpression() ||
                    parentPath.isJSXExpressionContainer() ||
                    (renderScope.hasBinding(path.node.name) && path.isReferencedIdentifier())) {
                    jsxReferencedIdentifiers.add(path.node);
                }
            },
            JSXAttribute(path) {
                const attr = path.node;
                const name = attr.name;
                let value;
                // @TODO: 使用作用域的方法全部替换，现在 cover 的情况太少
                if (t.isJSXElement(attr.value)) {
                    throw utils_1.codeFrameError(attr.loc, 'JSX 参数不支持传入 JSX 元素');
                }
                else if (t.isJSXExpressionContainer(attr.value)) {
                    const expresion = attr.value.expression;
                    const expresionPath = path.get('value').get('expression');
                    if (t.isIdentifier(expresion)) {
                        jsxReferencedIdentifiers.add(expresion);
                        expresionPath.replaceWith(t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), expresion));
                    }
                    else if (t.isMemberExpression(expresion)) {
                        if (!t.isThisExpression(expresion.object) && !isContainThis(expresion)) {
                            if (t.isIdentifier(expresion.object)) {
                                jsxReferencedIdentifiers.add(expresion.object);
                            }
                            expresionPath.replaceWith(buildInternalSafeGet(babel_generator_1.default(expresion).code));
                        }
                        else if (isNeedClassMethodWrapper(name.name, expresion)) {
                            const id = `$anonymousPropsFunction_${uid()}`;
                            path.get('value').get('expression').replaceWith(t.memberExpression(t.thisExpression(), t.identifier(id)));
                            body.push(buildAnonymousClassMethod(expresion, id));
                        }
                    }
                    else {
                        expresionPath.traverse({
                            Identifier(path) {
                                const { parent, node } = path;
                                if (!t.isMemberExpression(parent) && node.name !== constant_1.INTERNAL_SAFE_GET) {
                                    jsxReferencedIdentifiers.add(node);
                                    path.replaceWith(t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), node));
                                }
                            },
                            MemberExpression(path) {
                                const { parent, node } = path;
                                if ((parent === expresion) && !isContainThis(node)) {
                                    path.replaceWith(buildInternalSafeGet(babel_generator_1.default(node).code));
                                }
                            }
                        });
                    }
                    value = attr.value.expression;
                }
                else {
                    value = attr.value;
                }
                attrObj.push(t.objectProperty(t.identifier(name.name), value));
            }
        });
        properties.push(t.objectMethod('method', t.identifier(name), [], t.blockStatement([t.returnStatement(t.objectExpression(attrObj))])));
    });
    // console.log(jsxReferencedIdentifiers)
    return t.classProperty(t.identifier('$props'), t.objectExpression(properties));
}
function findImportedName(name) {
    return utils_1.isNumeric(name.slice(-1)) && name.slice(-2)[0] === '_'
        ? name.slice(0, name.length - 2)
        : name;
}
function build$ComponentsProperty(importedJSXElement, loopComponents) {
    const properties = [];
    for (const name of importedJSXElement.values()) {
        properties.push(t.objectProperty(t.identifier(name), t.identifier(findImportedName(name))));
    }
    loopComponents.forEach(lc => {
        lc.forEach(c => {
            if (!properties.some(p => t.isIdentifier(p.key) && p.key.name === c.name)) {
                properties.push(t.objectProperty(t.identifier(c.name), t.identifier(findImportedName(c.name))));
            }
        });
    });
    return t.classProperty(t.identifier('$components'), t.objectExpression(properties));
}
//# sourceMappingURL=class.js.map