"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const babel_traverse_1 = require("babel-traverse");
const babel_generator_1 = require("babel-generator");
const class_1 = require("./class");
const html_1 = require("html");
const utils_1 = require("./utils");
const t = require("babel-types");
const constant_1 = require("./constant");
const babel_core_1 = require("babel-core");
function transform(options) {
    const code = options.code;
    utils_1.setting.sourceCode = code;
    // babel-traverse 无法生成 Hub
    // 导致 Path#getSource|buildCodeFrameError 都无法直接使用
    // 原因大概是 babylon.parse 没有生成 File 实例导致 scope 和 path 原型上都没有 `file`
    // 将来升级到 babel@7 可以直接用 parse 而不是 transform
    const ast = babel_core_1.transform(code, {
        parserOpts: {
            sourceType: 'module',
            plugins: [
                'typescript',
                'classProperties',
                'jsx',
                'trailingFunctionCommas',
                'asyncFunctions',
                'exponentiationOperator',
                'asyncGenerators',
                'objectRestSpread',
                'decorators'
            ]
        }
    }).ast;
    // transformFromAst(ast, code)
    let result;
    const componentSourceMap = new Map();
    let mainClass;
    let storeName;
    let renderMethod;
    babel_traverse_1.default(ast, {
        ClassDeclaration(path) {
            mainClass = path;
        },
        ClassMethod(path) {
            if (t.isIdentifier(path.node.key) && path.node.key.name === 'render') {
                renderMethod = path;
            }
        },
        AwaitExpression(path) {
            const isAsyncImported = ast.program.body.some(statement => {
                return t.isImportDeclaration(statement) && statement.source.value === constant_1.ASYNC_PACKAGE_NAME;
            });
            if (!isAsyncImported) {
                ast.program.body.unshift(t.importDeclaration([], t.stringLiteral(constant_1.ASYNC_PACKAGE_NAME)));
            }
        },
        JSXOpeningElement(path) {
            const { name } = path.node.name;
            if (name === 'Provider') {
                const modules = path.scope.getAllBindings('module');
                const providerBinding = Object.values(modules).some((m) => m.identifier.name === 'Provider');
                if (providerBinding) {
                    path.node.name = t.jSXIdentifier('View');
                    const store = path.node.attributes.find(attr => attr.name.name === 'store');
                    if (store && t.isJSXExpressionContainer(store.value) && t.isIdentifier(store.value.expression)) {
                        storeName = store.value.expression.name;
                    }
                    path.node.attributes = [];
                }
            }
        },
        ImportDeclaration(path) {
            const source = path.node.source.value;
            const names = [];
            if (source === constant_1.TARO_PACKAGE_NAME) {
                path.node.specifiers.push(t.importSpecifier(t.identifier(constant_1.INTERNAL_SAFE_GET), t.identifier(constant_1.INTERNAL_SAFE_GET)), t.importSpecifier(t.identifier(constant_1.INTERNAL_DYNAMIC), t.identifier(constant_1.INTERNAL_DYNAMIC)));
            }
            if (source === constant_1.REDUX_PACKAGE_NAME) {
                path.node.specifiers.forEach((s, index, specs) => {
                    if (s.local.name === 'Provider') {
                        specs.splice(index, 1);
                        specs.push(t.importSpecifier(t.identifier('setStore'), t.identifier('setStore')));
                    }
                });
            }
            path.traverse({
                ImportDefaultSpecifier(path) {
                    const name = path.node.local.name;
                    constant_1.DEFAULT_Component_SET.has(name) || names.push(name);
                },
                ImportSpecifier(path) {
                    const name = path.node.local.name;
                    constant_1.DEFAULT_Component_SET.has(name) || names.push(name);
                }
            });
            componentSourceMap.set(source, names);
        }
    });
    const storeBinding = mainClass.scope.getBinding(storeName);
    if (storeBinding) {
        const statementPath = storeBinding.path.getStatementParent();
        if (statementPath) {
            ast.program.body.forEach((node, index, body) => {
                if (node === statementPath.node) {
                    body.splice(index + 1, 0, t.expressionStatement(t.callExpression(t.identifier('setStore'), [
                        t.identifier('store')
                    ])));
                }
            });
        }
    }
    if (options.isApp) {
        renderMethod.remove();
        return { ast };
    }
    result = new class_1.Transformer(mainClass, options.isRoot, componentSourceMap, options.path).result;
    result.code = babel_generator_1.default(ast).code;
    result.ast = ast;
    result.template = html_1.prettyPrint(result.template);
    return result;
}
exports.default = transform;
//# sourceMappingURL=index.js.map