"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const t = require("babel-types");
const utils_1 = require("./utils");
const jsx_1 = require("./jsx");
const constant_1 = require("./constant");
const babel_generator_1 = require("babel-generator");
const template = require('babel-template');
function isContainStopPropagation(path) {
    let bool = false;
    path.traverse({
        Identifier(p) {
            if (p.node.name === 'stopPropagation' &&
                p.parentPath.parentPath.isCallExpression()) {
                bool = true;
            }
        }
    });
    return bool;
}
function buildConstVariableDeclaration(variableName, expresion) {
    return t.variableDeclaration('const', [
        t.variableDeclarator(t.identifier(variableName), expresion)
    ]);
}
function buildComponentPathDataset(path) {
    const parentPath = path.parentPath;
    if (parentPath.isJSXOpeningElement()) {
        parentPath.node.attributes.push(t.jSXAttribute(t.jSXIdentifier('data-component-path'), t.stringLiteral('{{$path}}')));
    }
}
function buildAssignState(pendingState, initState) {
    return t.expressionStatement(t.callExpression(t.memberExpression(t.identifier('Object'), t.identifier('assign')), [
        t.memberExpression(t.thisExpression(), t.identifier('state')),
        pendingState
    ]));
}
function copyStateToShalowData() {
    return t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), t.identifier('__data')), t.callExpression(t.memberExpression(t.identifier('Object'), t.identifier('assign')), [
        t.objectExpression([]),
        t.memberExpression(t.thisExpression(), t.identifier('state'))
    ])));
}
function setTemplate(name, path, templates) {
    const parentPath = path.parentPath;
    const jsxChildren = parentPath.findParent(p => p.isJSXElement());
    if (name && !jsxChildren) {
        templates.set(name, path.node);
    }
}
function isContainFunction(p) {
    let bool = false;
    p.traverse({
        CallExpression() {
            bool = true;
        }
    });
    return bool;
}
function parseRender(renderPath, methods, initState, isRoot, instanceName, referencedIdentifiers, customComponentNames) {
    const body = renderPath.get('body');
    const renderScope = body.scope;
    // const referencedIdentifiers = new Set<t.Identifier>()
    const classProperties = new Set();
    const templates = new Map();
    const jsxDeclarations = new Set();
    const loopScopes = new Set();
    const anonymousState = new Set();
    const returnedPaths = [];
    const [, error] = renderPath.node.body.body.filter(s => t.isReturnStatement(s));
    if (error) {
        throw utils_1.codeFrameError(error.loc, 'render 函数顶级作用域暂时只支持一个 return');
    }
    let finalReturnElement;
    // replaceReturnNullToViewElement(body)
    body.traverse({
        NullLiteral(path) {
            const statementParent = path.getStatementParent();
            if (statementParent && statementParent.isReturnStatement()) {
                path.replaceWith(t.jSXElement(t.jSXOpeningElement(t.jSXIdentifier('View'), []), undefined, [], true));
            }
        },
        JSXElement(jsxElementPath) {
            const parentNode = jsxElementPath.parent;
            const parentPath = jsxElementPath.parentPath;
            const isFinalReturn = jsxElementPath.getFunctionParent().isClassMethod();
            const isJSXChildren = t.isJSXElement(parentNode);
            if (!isJSXChildren) {
                let statementParent = jsxElementPath.getStatementParent();
                if (!(statementParent.isVariableDeclaration() ||
                    statementParent.isExpressionStatement())) {
                    statementParent = statementParent.findParent(s => s.isVariableDeclaration() || s.isExpressionStatement());
                }
                jsxDeclarations.add(statementParent);
                if (t.isVariableDeclarator(parentNode)) {
                    if (statementParent) {
                        const name = utils_1.findIdentifierFromStatement(statementParent.node);
                        // setTemplate(name, path, templates)
                        name && templates.set(name, jsxElementPath.node);
                    }
                }
                else if (t.isLogicalExpression(parentNode)) {
                    const { left, operator } = parentNode;
                    if (operator === '&&') {
                        if (t.isExpression(left)) {
                            utils_1.newJSXIfAttr(jsxElementPath.node, left);
                            parentPath.replaceWith(jsxElementPath.node);
                            if (statementParent) {
                                const name = utils_1.findIdentifierFromStatement(statementParent.node);
                                setTemplate(name, jsxElementPath, templates);
                                // name && templates.set(name, path.node)
                            }
                        }
                    }
                }
                else if (t.isConditionalExpression(parentNode)) {
                    const { test, consequent, alternate } = parentNode;
                    const block = jsx_1.buildBlockElement();
                    if (t.isJSXElement(consequent) && t.isLiteral(alternate)) {
                        const { value, confident } = parentPath.get('alternate').evaluate();
                        if (confident && !value) {
                            utils_1.newJSXIfAttr(block, test);
                            block.children = [jsxElementPath.node];
                            // newJSXIfAttr(jsxElementPath.node, test)
                            parentPath.replaceWith(block);
                            if (statementParent) {
                                const name = utils_1.findIdentifierFromStatement(statementParent.node);
                                setTemplate(name, jsxElementPath, templates);
                                // name && templates.set(name, path.node)
                            }
                        }
                    }
                    else if (t.isLiteral(consequent) && t.isJSXElement(consequent)) {
                        if (t.isNullLiteral(consequent)) {
                            utils_1.newJSXIfAttr(block, utils_1.reverseBoolean(test));
                            // newJSXIfAttr(jsxElementPath.node, reverseBoolean(test))
                            parentPath.replaceWith(block);
                            if (statementParent) {
                                const name = utils_1.findIdentifierFromStatement(statementParent.node);
                                setTemplate(name, jsxElementPath, templates);
                                // name && templates.set(name, path.node)
                            }
                        }
                    }
                    else if (t.isJSXElement(consequent) && t.isJSXElement(alternate)) {
                        const block2 = jsx_1.buildBlockElement();
                        block.children = [consequent];
                        utils_1.newJSXIfAttr(block, test);
                        jsx_1.setJSXAttr(block2, 'wx:else');
                        block2.children = [alternate];
                        const parentBlock = jsx_1.buildBlockElement();
                        parentBlock.children = [block, block2];
                        parentPath.replaceWith(parentBlock);
                        if (statementParent) {
                            const name = utils_1.findIdentifierFromStatement(statementParent.node);
                            setTemplate(name, jsxElementPath, templates);
                        }
                    }
                    else {
                        // console.log('todo')
                    }
                }
                else if (t.isReturnStatement(parentNode)) {
                    if (!isFinalReturn) {
                        const caller = parentPath.findParent(p => p.isCallExpression());
                        if (caller.isCallExpression()) {
                            const callee = caller.node.callee;
                            if (t.isMemberExpression(callee) &&
                                t.isIdentifier(callee.property) &&
                                callee.property.name === 'map') {
                                let ary = callee.object;
                                // const blockStatementPath = parentPath.findParent(p => p.isBlockStatement()) as NodePath<t.BlockStatement>
                                if (t.isCallExpression(ary) || isContainFunction(caller.get('callee').get('object'))) {
                                    const variableName = `anonymousState_${anonymousState.size}`;
                                    anonymousState.add(buildConstVariableDeclaration(variableName, ary));
                                    ary = t.identifier(variableName);
                                }
                                jsx_1.setJSXAttr(jsxElementPath.node, 'wx:for', t.jSXExpressionContainer(ary));
                                const [func] = caller.node.arguments;
                                if (t.isFunctionExpression(func) ||
                                    t.isArrowFunctionExpression(func)) {
                                    const [items, index] = func.params;
                                    if (t.isIdentifier(items)) {
                                        jsx_1.setJSXAttr(jsxElementPath.node, 'wx:for-item', t.stringLiteral(items.name));
                                        loopScopes.add(items.name);
                                    }
                                    if (t.isIdentifier(index)) {
                                        jsx_1.setJSXAttr(jsxElementPath.node, 'wx:for-index', t.stringLiteral(index.name));
                                        loopScopes.add(index.name);
                                    }
                                    caller.replaceWith(jsxElementPath.node);
                                    if (statementParent) {
                                        const name = utils_1.findIdentifierFromStatement(statementParent.node);
                                        // setTemplate(name, path, templates)
                                        name && templates.set(name, jsxElementPath.node);
                                    }
                                }
                            }
                        }
                    }
                    else {
                        const ifStatement = parentPath.findParent(p => p.isIfStatement());
                        const blockStatement = parentPath.findParent(p => p.isBlockStatement());
                        const block = finalReturnElement || jsx_1.buildBlockElement();
                        if (utils_1.isBlockIfStatement(ifStatement, blockStatement)) {
                            const { test, alternate, consequent } = ifStatement.node;
                            if (alternate === blockStatement.node) {
                                throw utils_1.codeFrameError(parentNode.loc, '不必要的 else 分支，请遵从 ESLint consistent-return: https://eslint.org/docs/rules/consistent-return');
                            }
                            else if (consequent === blockStatement.node) {
                                const parentIfStatement = ifStatement.findParent(p => p.isIfStatement());
                                if (parentIfStatement) {
                                    jsx_1.setJSXAttr(jsxElementPath.node, 'wx:elif', t.jSXExpressionContainer(test));
                                }
                                else {
                                    utils_1.newJSXIfAttr(jsxElementPath.node, test);
                                }
                            }
                        }
                        else if (block.children.length !== 0) {
                            jsx_1.setJSXAttr(jsxElementPath.node, 'wx:else');
                        }
                        block.children.push(jsxElementPath.node);
                        finalReturnElement = block;
                        returnedPaths.push(parentPath);
                    }
                }
                else if (t.isArrowFunctionExpression(parentNode)) {
                    //
                }
                else if (t.isAssignmentExpression(parentNode)) {
                    if (t.isIdentifier(parentNode.left)) {
                        const name = parentNode.left.name;
                        const bindingNode = renderScope.getOwnBinding(name).path.node;
                        const block = templates.get(name) || jsx_1.buildBlockElement();
                        if (utils_1.isEmptyDeclarator(bindingNode)) {
                            const ifStatement = parentPath.findParent(p => p.isIfStatement());
                            const blockStatement = parentPath.findParent(p => p.isBlockStatement());
                            if (utils_1.isBlockIfStatement(ifStatement, blockStatement)) {
                                const { test, alternate, consequent } = ifStatement.node;
                                if (alternate === blockStatement.node) {
                                    jsx_1.setJSXAttr(jsxElementPath.node, 'wx:else');
                                }
                                else if (consequent === blockStatement.node) {
                                    const parentIfStatement = ifStatement.findParent(p => p.isIfStatement());
                                    if (parentIfStatement && parentIfStatement.get('alternate') === ifStatement) {
                                        jsx_1.setJSXAttr(jsxElementPath.node, 'wx:elif', t.jSXExpressionContainer(test));
                                    }
                                    else {
                                        if (parentIfStatement) {
                                            utils_1.newJSXIfAttr(block, parentIfStatement.node.test);
                                        }
                                        utils_1.newJSXIfAttr(jsxElementPath.node, test);
                                    }
                                }
                                block.children.push(jsxElementPath.node);
                                // setTemplate(name, path, templates)
                                name && templates.set(name, block);
                            }
                        }
                        else {
                            throw utils_1.codeFrameError(jsxElementPath.node.loc, '请将 JSX 赋值表达式初始化为 null，然后再进行 if 条件表达式赋值。');
                        }
                    }
                }
                else if (!t.isJSXElement(parentNode)) {
                    // throwError(path, '考虑只对 JSX 元素赋值一次。')
                }
            }
            // handle jsx attrs
            const openingElementPath = jsxElementPath.get('openingElement');
            openingElementPath.traverse({
                JSXExpressionContainer(path) {
                    const expression = path.get('expression');
                    if (expression.isStringLiteral()) {
                        path.replaceWith(expression);
                    }
                    else if (expression.isCallExpression()) {
                        const node = expression.node;
                        if (t.isMemberExpression(node.callee) &&
                            t.isIdentifier(node.callee.property) &&
                            node.callee.property.name === 'bind') {
                            const JSXElement = path.findParent(p => p.isJSXElement())
                                .node;
                            // const JSXAttribute = path.findParent(p => p.isJSXAttribute())
                            let bindCalleeName = null;
                            if (t.isIdentifier(node.callee.object)) {
                                bindCalleeName = node.callee.object.name;
                            }
                            else if (t.isMemberExpression(node.callee.object)) {
                                if (t.isIdentifier(node.callee.object.property)) {
                                    bindCalleeName = node.callee.object.property.name;
                                }
                            }
                            if (bindCalleeName !== null) {
                                node.arguments.forEach((arg, index) => {
                                    const argName = utils_1.getArgumentName(arg);
                                    const isLiteral = t.isLiteral(arg);
                                    if (index === 0) {
                                        jsx_1.setJSXAttr(JSXElement, `data-event-${bindCalleeName}-scope`, t.stringLiteral(argName));
                                    }
                                    else {
                                        if (path.scope.hasBinding(argName) && t.isIdentifier(arg)) {
                                            referencedIdentifiers.add(arg);
                                        }
                                        jsx_1.setJSXAttr(JSXElement, `data-event-${bindCalleeName}-arg-${utils_1.toLetters(index)}`, isLiteral
                                            ? t.stringLiteral(argName)
                                            : t.jSXExpressionContainer(t.identifier(argName)));
                                    }
                                });
                                jsx_1.setJSXAttr(JSXElement, 'data-component-path', t.stringLiteral('{{$path}}'));
                                expression.replaceWith(t.stringLiteral(`${!isRoot ? `${instanceName}__` : ''}${bindCalleeName}`));
                            }
                        }
                    }
                    else if (utils_1.isComplexExpression(expression)) {
                        const variableName = `anonymousState_${anonymousState.size}`;
                        anonymousState.add(buildConstVariableDeclaration(variableName, expression.node));
                        expression.replaceWith(t.identifier(variableName));
                    }
                },
                MemberExpression(path) {
                    const id = utils_1.findFirstIdentifierFromMemberExpression(path.node);
                    const bindId = renderScope.getOwnBindingIdentifier(id.name);
                    if (bindId) {
                        referencedIdentifiers.add(bindId);
                    }
                },
                JSXAttribute(path) {
                    const { name, value } = path.node;
                    let eventShouldBeCatched = false;
                    if (t.isJSXIdentifier(name)) {
                        if (!isRoot && name.name.startsWith('on')) {
                            buildComponentPathDataset(path);
                            const jsxElementPath = path.parentPath.parentPath;
                            if (jsxElementPath.isJSXElement()) {
                                const alreadySet = jsxElementPath.node.openingElement.attributes.some(attr => {
                                    return t.isJSXIdentifier(attr.name) && attr.name.name === 'data-component-class';
                                });
                                if (!alreadySet) {
                                    jsx_1.setJSXAttr(jsxElementPath.node, 'data-component-class', t.stringLiteral(instanceName));
                                }
                            }
                        }
                        if (name.name === 'key') {
                            path.get('name').replaceWith(t.jSXIdentifier('wx:key'));
                        }
                        else if (name.name.startsWith('on')) {
                            if (t.isJSXExpressionContainer(value)) {
                                let methodName;
                                if (t.isIdentifier(value.expression) ||
                                    t.isJSXIdentifier(value.expression)) {
                                    methodName = value.expression.name;
                                }
                                else if (t.isMemberExpression(value.expression) &&
                                    t.isIdentifier(value.expression.property)) {
                                    const { code } = babel_generator_1.default(value.expression);
                                    const ids = code.split('.');
                                    if (ids[0] === 'this' && ids[1] === 'props' && ids[2]) {
                                        // const method = ids
                                    }
                                    else {
                                        methodName = value.expression.property.name;
                                    }
                                }
                                else if (t.isCallExpression(value.expression) &&
                                    t.isMemberExpression(value.expression.callee) &&
                                    t.isIdentifier(value.expression.callee.object)) {
                                    methodName = value.expression.callee.object.name;
                                }
                                else if (t.isCallExpression(value.expression) &&
                                    t.isMemberExpression(value.expression.callee) &&
                                    t.isMemberExpression(value.expression.callee.object) &&
                                    t.isIdentifier(value.expression.callee.property) &&
                                    value.expression.callee.property.name === 'bind' &&
                                    t.isIdentifier(value.expression.callee.object.property)) {
                                    methodName = value.expression.callee.object.property.name;
                                }
                                else {
                                    throw utils_1.codeFrameError(path.node.loc, '当 props 为事件时(props name 以 `on` 开头)，只能传入一个 this 作用域下的函数。');
                                }
                                if (methods.has(methodName)) {
                                    const method = methods.get(methodName);
                                    if (t.isIdentifier(method.node.key)) {
                                        method.node.key = t.identifier('__event_' + `${isRoot ? '' : `${instanceName}__`}` + methodName);
                                    }
                                    if (!isRoot && !babel_generator_1.default(value.expression).code.includes('.bind')) {
                                        path.node.value = t.stringLiteral(`${instanceName}__${methodName}`);
                                    }
                                    eventShouldBeCatched = isContainStopPropagation(method);
                                }
                            }
                            if (t.isJSXIdentifier(jsxElementPath.node.openingElement.name)
                                && constant_1.DEFAULT_Component_SET.has(jsxElementPath.node.openingElement.name.name)) {
                                let transformName = `${eventShouldBeCatched ? 'catch' : 'bind'}`
                                    + name.name.slice(2, name.name.length).toLowerCase();
                                if (name.name === 'onClick') {
                                    transformName = eventShouldBeCatched ? 'catchtap' : 'bindtap';
                                }
                                path.node.name = t.jSXIdentifier(transformName);
                            }
                            // let transformName = `${eventShouldBeCatched ? 'catch' : 'bind'}` + name.name.slice(2, name.name.length)
                            // transformName = eventShouldBeCatched
                            //   ? CATCH_EVENT_MAP.get(name.name)!
                            //   : BIND_EVENT_MAP.get(name.name)!
                        }
                    }
                },
                Identifier(path) {
                    const parentPath = path.parentPath;
                    if (parentPath.isConditionalExpression() ||
                        parentPath.isLogicalExpression() ||
                        parentPath.isJSXExpressionContainer() ||
                        (renderScope.hasBinding(path.node.name))) {
                        referencedIdentifiers.add(path.node);
                    }
                },
                ArrowFunctionExpression(path) {
                    const uid = path.scope.generateUid('_anonymous_function_');
                    const c = t.classProperty(t.identifier(uid), path.node);
                    classProperties.add(c);
                }
            });
        },
        JSXExpressionContainer(path) {
            // todo
            if (!t.isJSXAttribute(path.parent)) {
                path.traverse({
                    Identifier(path) {
                        const parentPath = path.parentPath;
                        if (parentPath.isConditionalExpression() ||
                            parentPath.isLogicalExpression() ||
                            path.isReferencedIdentifier()) {
                            const name = path.node.name;
                            if (Object.keys(renderScope.getAllBindings()).includes(name)) {
                                referencedIdentifiers.add(path.node);
                            }
                            if (templates.has(name)) {
                                path.replaceWith(jsx_1.buildRefTemplate(name));
                            }
                        }
                    }
                });
                const expression = path.get('expression');
                if (utils_1.isComplexExpression(expression)) {
                    const variableName = `anonymousState_${anonymousState.size}`;
                    anonymousState.add(buildConstVariableDeclaration(variableName, expression.node));
                    expression.replaceWith(t.identifier(variableName));
                }
            }
        }
    });
    // TODO: 这里还得确定一下 template 的作用域，可能需要把 template 全包起来
    let outputTemplate = !isRoot
        ? jsx_1.generateHTMLTemplate(finalReturnElement, instanceName)
        : jsx_1.parseJSXElement(finalReturnElement);
    templates.forEach((template, name) => {
        outputTemplate += ` ${jsx_1.generateHTMLTemplate(template, name)}`;
    });
    jsxDeclarations.forEach(d => d && d.remove());
    returnedPaths.forEach(p => p.remove());
    const usedState = Array.from(new Set(Array.from(referencedIdentifiers)
        .map(i => i.name)))
        .filter(i => {
        return !methods.has(i);
    })
        .filter(i => !loopScopes.has(i))
        .filter(i => !initState.has(i))
        .filter(i => !templates.has(i));
    const classPath = renderPath.findParent(p => p.isClassDeclaration());
    classPath.node.body.body.unshift(t.classProperty(t.identifier('$usedState'), t.arrayExpression(usedState
        .filter(s => !loopScopes.has(s.split('.')[0]))
        .concat(Array.from(customComponentNames))
        .map(s => t.stringLiteral(s)))));
    const pendingState = t.objectExpression(usedState.filter(i => !i.includes('.'))
        .filter(i => i !== constant_1.MAP_CALL_ITERATOR)
        .map(i => t.objectProperty(t.identifier(i), t.identifier(i))));
    renderPath.node.body.body = renderPath.node.body.body.concat(Array.from(anonymousState), buildAssignState(pendingState, initState), copyStateToShalowData(), t.returnStatement(t.memberExpression(t.thisExpression(), t.identifier('state'))));
    body.traverse({
        ThisExpression(path) {
            const property = path.getSibling('property');
            if (property.isIdentifier({ name: 'state' })) {
                property.replaceWith(t.identifier('__state'));
            }
        }
    });
    body.insertAfter(template(`
const __state = this.__state;
delete this.__state;
return __state;`)());
    renderPath.node.body.body.unshift(template(`this.__state = arguments[0] || this.state;`)());
    if (t.isIdentifier(renderPath.node.key)) {
        renderPath.node.key.name = '_createData';
    }
    return outputTemplate;
}
exports.parseRender = parseRender;
//# sourceMappingURL=render.js.map